// SDL_ph_video.c
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#ifdef MTRR_SUPPORT
#include <asm/mtrr.h>
#include <sys/fcntl.h>
#endif

#include "SDL.h"
#include "SDL_error.h"
#include "SDL_timer.h"
#include "SDL_thread.h"
#include "SDL_video.h"
#include "SDL_mouse.h"
#include "SDL_endian.h"
#include "SDL_sysvideo.h"
#include "SDL_pixels_c.h"
#include "SDL_events_c.h"
#include "SDL_ph_video.h"
#include "SDL_ph_modes_c.h"
#include "SDL_ph_image_c.h"
#include "SDL_ph_events_c.h"
#include "SDL_ph_mouse_c.h"
#include "blank_cursor.h"

static int ph_VideoInit(_THIS, SDL_PixelFormat *vformat);
static SDL_Surface *ph_SetVideoMode(_THIS, SDL_Surface *current,
                int width, int height, int bpp, Uint32 flags);
static int ph_SetColors(_THIS, int firstcolor, int ncolors);
static void ph_VideoQuit(_THIS);
static void ph_DeleteDevice(SDL_VideoDevice *device);

static int ph_Available(void)
{
        return 1;
}

static SDL_VideoDevice *ph_CreateDevice(int devindex)
{
    SDL_VideoDevice *device;

    /* Initialize all variables that we clean on shutdown */
    device = (SDL_VideoDevice *)malloc(sizeof(SDL_VideoDevice));
    if ( device ) {
        memset(device, 0, (sizeof *device));
        device->hidden = (struct SDL_PrivateVideoData *)
                malloc((sizeof *device->hidden));
        device->gl_data = NULL;
    }
    if ( (device == NULL) || (device->hidden == NULL) ) {
        SDL_OutOfMemory();
        ph_DeleteDevice(device);
        return(0);
    }
    memset(device->hidden, 0, (sizeof *device->hidden));

    /* Set the driver flags */
    device->handles_any_size = 1;

    /* Set the function pointers */
    device->VideoInit = ph_VideoInit;
    device->ListModes = ph_ListModes;
    device->SetVideoMode = ph_SetVideoMode;
    device->ToggleFullScreen =  NULL; //ph_ToggleFullScreen;
    device->UpdateMouse = NULL;	
    device->SetColors = ph_SetColors;
    device->UpdateRects = NULL;
    device->VideoQuit = ph_VideoQuit;
    device->AllocHWSurface = ph_AllocHWSurface;
    device->CheckHWBlit = NULL;
    device->FillHWRect = NULL;
    device->SetHWColorKey = NULL;
    device->SetHWAlpha = NULL;
    device->LockHWSurface = ph_LockHWSurface;
    device->UnlockHWSurface = ph_UnlockHWSurface;
    device->FlipHWSurface = ph_FlipHWSurface;
    device->FreeHWSurface = ph_FreeHWSurface;
    device->SetIcon = NULL;
    device->SetCaption = NULL;
    device->IconifyWindow = NULL;
    device->GrabInput = NULL;
    device->GetWMInfo = NULL;
    device->FreeWMCursor = ph_FreeWMCursor;
    device->CreateWMCursor = ph_CreateWMCursor;
    device->ShowWMCursor = ph_ShowWMCursor;
    device->WarpWMCursor = ph_WarpWMCursor;
    device->CheckMouseMode = ph_CheckMouseMode;
    device->InitOSKeymap = ph_InitOSKeymap;
    device->PumpEvents = ph_PumpEvents;

    device->free = ph_DeleteDevice;

    return device;
}

VideoBootStrap X11_bootstrap = {
        "photon", "QNX Photon video output",
	ph_Available, ph_CreateDevice
};

static void ph_DeleteDevice(SDL_VideoDevice *device)
{
    if ( device ) {
        if ( device->hidden ) {
            free(device->hidden);
        }
        if ( device->gl_data ) {
            free(device->gl_data);
        }
        free(device);
    }
}

static int ph_VideoInit(_THIS, SDL_PixelFormat *vformat)
// TODO: add error checking
{
    /* Get the available video modes */
//    if(ph_GetVideoModes(this) < 0)
//        return -1;

	/* Create the blank cursor */
	SDL_BlankCursor = this->CreateWMCursor(this, blank_cdata, blank_cmask,
                BLANK_CWIDTH, BLANK_CHEIGHT,
                    BLANK_CHOTX, BLANK_CHOTY);

    return 0;
}

static SDL_Surface *ph_SetVideoMode(_THIS, SDL_Surface *current,
                int width, int height, int bpp, Uint32 flags)
{
    PhRegion_t region_info;
    PgDisplaySettings_t settings;
	PgVideoModeInfo_t mode_info;
    int mode, actual_width, actual_height;
	PtArg_t arg[5];
	PhDim_t dim;	

	actual_width = width;
	actual_height = height;

    /* Lock the event thread, in multi-threading environments */
    SDL_Lock_EventThread();

    /* Check the combination of flags we were passed */
    if ( flags & SDL_FULLSCREEN ) {
        /* Clear fullscreen flag if not supported */
        flags &= ~SDL_FULLSCREEN;
    }

    /* Initialize the window */
    if (flags & SDL_FULLSCREEN)
    {
        /* Get the video mode and set it */
        if (bpp == 0)
        {
            if (PgGetVideoMode( &settings ) < 0)
            {
                fprintf(stderr,"error: PgGetVideoMode failed\n");
            }
            if (PgGetVideoModeInfo(settings.mode, &mode_info) < 0)
            {
                fprintf(stderr,"error: PgGetVideoModeInfo failed\n");
            }
            bpp = mode_info.bits_per_pixel;
        }
        if (flags & SDL_ANYFORMAT)
        {
            if ((mode = get_mode_any_format(width, height, bpp)) == 0)
            {
                fprintf(stderr,"error: get_mode_any_format failed\n");
            }
        }
        else
        {
            if ((mode = get_mode(width, height, bpp)) == 0)
            {
                fprintf(stderr,"error: get_mode failed\n");
            }
        }
        settings.mode = mode;
        if (PgSetVideoMode( &settings ) < 0)
        {
            fprintf(stderr,"error: PgSetVideoMode failed\n");
        }

		/* Get the true height and width */
		if (PgGetVideoMode( &settings ) < 0)
        {
            fprintf(stderr,"error: PgGetVideoMode failed\n");
        }
        if (PgGetVideoModeInfo(settings.mode, &mode_info) < 0)
        {
            fprintf(stderr,"error: PgGetVideoModeInfo failed\n");
        }
		current_w = mode_info.width;
		current_h = mode_info.height;
		actual_width = current_w;
		actual_height = current_h;

        /* Begin direct mode */
        directContext=(PdDirectContext_t*)PdCreateDirectContext();
        if( !directContext ) fprintf(stderr, "error: Can't create direct context\n" );
            PdDirectStart( directContext );
    }

    /* Remove the cursor if in full screen mode */
    if (flags & SDL_FULLSCREEN)
    {
        region_info.cursor_type = Ph_CURSOR_NONE;
        region_info.rid=PtWidgetRid(window);
        PhRegionChange(Ph_REGION_CURSOR,0,&region_info,NULL,NULL);
    }

    /* Use offscreen memory iff SDL_HWSURFACE flag is set */
    if (flags & SDL_HWSURFACE)
    {
        Buff[0] = PdCreateOffscreenContext(0,0,0,Pg_OSC_MAIN_DISPLAY | Pg_OSC_CRTC_SAFE);
        if (flags & SDL_DOUBLEBUF)
        {
            Buff[1]=PdDupOffscreenContext(Buff[0],Pg_OSC_CRTC_SAFE | Pg_OSC_MEM_PAGE_ALIGN);
            PhDCSetCurrent( Buff[1] );
        }
        else
        {
            PhDCSetCurrent( Buff[0] );
        }
    }
    /* Set up the new mode framebuffer */
    if ( (current->w != actual_width) || (current->h != actual_height) ) {
        current->w = actual_width;
        current->h = actual_height;
        current->pitch = SDL_CalculatePitch(current);
		ph_ResizeImage(this, current, flags);
    }

    current->flags |= (flags&SDL_RESIZABLE);


	/* Initialize or resize the window */
    //PtSetArg(&arg[0], Pt_ARG_WINDOW_TITLE,"Q3",0);
    dim.w=actual_width;
    dim.h=actual_height;
    PtSetArg(&arg[0], Pt_ARG_DIM, &dim,0);
    PtSetArg(&arg[1], Pt_ARG_RESIZE_FLAGS, Pt_TRUE, Pt_RESIZE_XY_AS_REQUIRED);
    PtSetArg(&arg[2], Pt_ARG_WINDOW_STATE, Pt_TRUE, Ph_WM_STATE_ISFRONT |
                        Ph_WM_STATE_ISMAX |
                        Ph_WM_STATE_ISFOCUS);

    PtSetArg(&arg[3], Pt_ARG_WINDOW_RENDER_FLAGS,Pt_FALSE,~0);
    PtSetArg(&arg[4], Pt_ARG_WINDOW_MANAGED_FLAGS,Pt_TRUE,
                        Ph_WM_FFRONT |
                        Ph_WM_CLOSE |
                        Ph_WM_TOFRONT |
                        Ph_WM_CONSWITCH);
	if (window)
	{
		 PtSetParentWidget(NULL);
		 window = PtCreateWidget(PtWidget,NULL,5,arg);
	}
	else
	{
	    window=PtAppInit(&app, NULL, NULL, 5, arg);
	    PgSetDrawBufferSize(16 *1024);
	    PgSetRegion(PtWidgetRid(window));
	    PgSetClipping(0,NULL);
    	PtRealizeWidget(window);
	}

    SDL_Unlock_EventThread();

    /* We're done! */
    return(current);
}

static void ph_VideoQuit(_THIS)
{
	PdDirectStop( directContext );
	PdReleaseDirectContext( directContext );
	directContext=0;	
}

static int ph_SetColors(_THIS, int firstcolor, int ncolors)
{
	SDL_Color *colors;
	PgColor_t *in, *out;
	int i;
	int alloct_all = 1;

    	colors = this->screen->format->palette->colors;

	in = alloca( ncolors*sizeof(PgColor_t)  );
	if ( in == NULL  ) {
		return 0;
	}
	memset(in,0,ncolors*sizeof(PgColor_t));

        out = alloca( ncolors*sizeof(PgColor_t)  );
        if ( in == NULL  ) {
                return 0;
        }
	
	for (i=0;i<ncolors;i++)
	{
		in[i] |= colors[i].r<<16 ;
		in[i] |= colors[i].g<<8 ;
		in[i] |= colors[i].b ; 
	}

	if ( (this->screen->flags & SDL_HWPALETTE) == SDL_HWPALETTE ) 
	{
		if( PgSetPalette( in, 0, firstcolor, ncolors, Pg_PALSET_HARD, 0) < 0 )
		{
			//TODO: handle error
		}  
	}
	else 
	{
	        if ( PgColorMatch(ncolors, in, out) < 0 )
        	{
             	   //TODO: handle error
        	}
		for (i=0;i<ncolors;i++)
		{
			if (memcmp(&in[i],&out[i],sizeof(PgColor_t)))
			{
				alloct_all = 0;
			}
		}
	}
	return alloct_all;
}

static int ph_ResizeWindow(_THIS,
            SDL_Surface *screen, int w, int h, Uint32 flags)
{
	PhWindowEvent_t winevent;

	memset( &winevent, 0, sizeof(winevent) ); 
	winevent.event_f = Ph_WM_RESIZE;
	winevent.size.w = w;
	winevent.size.h = h;
	winevent.rid = PtWidgetRid( window );
	if (PtForwardWindowEvent( &winevent ) < 0)
	{
		fprintf(stderr,"error: PtForwardWindowEvent failed.\n");
	}
	current_w = w;
	current_h = h;
    return(0);
}
